"""
XML Parser.
"""
import dateparser as dateparser
import logging
import xmltodict

from georss_client.consts import XML_TAG_DC_DATE, XML_TAG_LAST_BUILD_DATE, \
    XML_TAG_PUB_DATE, XML_TAG_PUBLISHED, XML_TAG_UPDATED, XML_TAG_GEO_LAT, \
    XML_TAG_GEO_LONG, XML_TAG_GEORSS_POLYGON, XML_TAG_GML_POS_LIST, \
    XML_TAG_GML_POS, XML_TAG_GEORSS_POINT, XML_TAG_HEIGHT, XML_TAG_TTL, \
    XML_TAG_WIDTH, XML_TAG_RSS, XML_TAG_CHANNEL, XML_TAG_FEED
from georss_client.xml_parser.feed import Feed

_LOGGER = logging.getLogger(__name__)

DEFAULT_NAMESPACES = {
    'http://www.w3.org/2005/Atom': None,
    'http://purl.org/dc/elements/1.1/': 'dc',
    'http://www.georss.org/georss': 'georss',
    'http://www.w3.org/2003/01/geo/wgs84_pos#': 'geo',
    'http://www.opengis.net/gml': 'gml',
    'http://www.gdacs.org/': 'gdacs',
}
KEYS_DATE = [XML_TAG_DC_DATE, XML_TAG_LAST_BUILD_DATE, XML_TAG_PUB_DATE,
             XML_TAG_PUBLISHED, XML_TAG_UPDATED]
KEYS_FLOAT = [XML_TAG_GEO_LAT, XML_TAG_GEO_LONG]
KEYS_FLOAT_LIST = [XML_TAG_GEORSS_POLYGON, XML_TAG_GML_POS_LIST,
                   XML_TAG_GML_POS, XML_TAG_GEORSS_POINT]
KEYS_INT = [XML_TAG_HEIGHT, XML_TAG_TTL, XML_TAG_WIDTH]


class XmlParser:
    """Built-in XML parser."""

    def __init__(self, additional_namespaces=None):
        """Initialise the XML parser."""
        self._namespaces = DEFAULT_NAMESPACES
        if additional_namespaces:
            self._namespaces.update(additional_namespaces)

    def parse(self, xml):
        """Parse the provided xml."""
        if xml:

            def postprocessor(path, key, value):
                """Conduct type conversion for selected keys."""
                try:
                    if key in KEYS_DATE and value:
                        return key, dateparser.parse(value)
                    if key in KEYS_FLOAT and value:
                        return key, float(value)
                    if key in KEYS_FLOAT_LIST and value:
                        # Turn white-space separated list of numbers into
                        # list of floats.
                        coordinate_values = value.split()
                        point_coordinates = []
                        for i in range(0, len(coordinate_values)):
                            point_coordinates.append(
                                float(coordinate_values[i]))
                        return key, point_coordinates
                    if key in KEYS_INT and value:
                        return key, int(value)
                except (ValueError, TypeError) as error:
                    _LOGGER.warning("Unable to process (%s/%s): %s",
                                    key, value, error)
                return key, value

            parsed_dict = xmltodict.parse(
                xml, process_namespaces=True, namespaces=self._namespaces,
                postprocessor=postprocessor)

            if XML_TAG_RSS in parsed_dict:
                rss = parsed_dict.get(XML_TAG_RSS)
                if XML_TAG_CHANNEL in rss:
                    channel = rss.get(XML_TAG_CHANNEL)
                    return Feed(channel)
            if XML_TAG_FEED in parsed_dict:
                feed = parsed_dict.get(XML_TAG_FEED)
                return Feed(feed)

        return None



"""Test for the INGV Centro Nazionale Terremoti feed."""
import datetime
import unittest
from unittest import mock

from georss_client import UPDATE_OK
from georss_client.ingv_centro_nazionale_terremoti_feed import \
    IngvCentroNazionaleTerremotiFeed
from tests.utils import load_fixture


class TestIngvCentroNazionaleTerremotiFeed(unittest.TestCase):
    """Test the INGV Centro Nazionale Terremoti feed."""

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (40.84, 14.25)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('ingv_centro_nazionale_terremoti_feed.xml')

        feed = IngvCentroNazionaleTerremotiFeed(home_coordinates)
        assert repr(feed) == "<IngvCentroNazionaleTerremotiFeed(home=" \
                             "(40.84, 14.25), url=http://cnt.rm.ingv.it/" \
                             "feed/atom/all_week, radius=None, " \
                             "magnitude=None)>"
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 2

        feed_entry = entries[0]
        assert feed_entry.title == "2018-10-06 10:21:33 UTC - Magnitude(ML)" \
                                   " 2.3 - 1 km NE Biancavilla (CT)"
        assert feed_entry.external_id == "smi:webservices.ingv.it/fdsnws/" \
                                         "event/1/query?eventId=1234"
        assert feed_entry.coordinates == (37.654, 14.878)
        self.assertAlmostEqual(feed_entry.distance_to_home, 358.4, 1)
        assert feed_entry.published \
            == datetime.datetime(2018, 10, 6, 8, 0,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.updated \
            == datetime.datetime(2018, 10, 6, 8, 30,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.region == "1 km NE Biancavilla (CT)"
        assert feed_entry.magnitude == 2.3
        assert feed_entry.attribution == "Istituto Nazionale di Geofisica " \
                                         "e Vulcanologia"
        assert repr(feed_entry) == "<IngvCentroNazionaleTerremotiFeedEntry" \
                                   "(id=smi:webservices.ingv.it/fdsnws/" \
                                   "event/1/query?eventId=1234)>"

        feed_entry = entries[1]
        assert feed_entry.title == "2018-10-06 09:14:11 UTC - Magnitude(ML)" \
                                   " 0.7 - 1 km NE Norcia (PG)"
        self.assertIsNone(feed_entry.published)

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_with_category(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (40.84, 14.25)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('ingv_centro_nazionale_terremoti_feed.xml')

        feed = IngvCentroNazionaleTerremotiFeed(home_coordinates,
                                                filter_minimum_magnitude=2.0)
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 1

        feed_entry = entries[0]
        assert feed_entry.title == "2018-10-06 10:21:33 UTC - Magnitude(ML)" \
                                   " 2.3 - 1 km NE Biancavilla (CT)"
        assert feed_entry.external_id == "smi:webservices.ingv.it/fdsnws/" \
                                         "event/1/query?eventId=1234"



"""Test utilities."""
import os


def load_fixture(filename):
    """Load a fixture."""
    path = os.path.join(os.path.dirname(__file__), 'fixtures', filename)
    with open(path, encoding='utf-8') as fptr:
        return fptr.read()



"""Test for the IGN Instituto Geográfico Nacional Sismología feed."""
import datetime
import unittest
from unittest import mock

from georss_client import UPDATE_OK
from georss_client.ign_sismologia_feed import IgnSismologiaFeed, ATTRIBUTION
from tests.utils import load_fixture


class TestIgnSismologiaFeed(unittest.TestCase):
    """Test the IGN Instituto Geográfico Nacional Sismología feed."""

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (40.38, -3.72)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('ign_sismologia_feed.xml')

        feed = IgnSismologiaFeed(home_coordinates)
        assert repr(feed) == "<IgnSismologiaFeed(home=" \
                             "(40.38, -3.72), url=http://www.ign.es/ign/" \
                             "RssTools/sismologia.xml, radius=None, " \
                             "magnitude=None)>"
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 3

        feed_entry = entries[0]
        assert feed_entry.title == "-Info.terremoto: 18/03/2019 19:34:55"
        assert feed_entry.external_id == "http://www.ign.es/web/ign/portal/" \
                                         "sis-catalogo-terremotos/-/" \
                                         "catalogo-terremotos/" \
                                         "detailTerremoto?evid=es2019dfcfp"
        assert feed_entry.coordinates == (27.62, -18.0859)
        self.assertAlmostEqual(feed_entry.distance_to_home, 1935.7, 1)
        assert feed_entry.published \
            == datetime.datetime(2019, 3, 18, 19, 34, 55)
        assert feed_entry.region == "SW EL PINAR DE EL HIERRO.IHI"
        assert feed_entry.magnitude == 3.1
        assert feed_entry.attribution == ATTRIBUTION
        assert feed_entry.image_url == "http://www.ign.es/web/resources/" \
                                       "sismologia/www/" \
                                       "dir_images_terremotos/detalle/" \
                                       "es2019dfcfp.gif"
        assert repr(feed_entry) == "<IgnSismologiaFeedEntry" \
                                   "(id=http://www.ign.es/web/ign/portal/" \
                                   "sis-catalogo-terremotos/-/" \
                                   "catalogo-terremotos/detailTerremoto?" \
                                   "evid=es2019dfcfp)>"

        feed_entry = entries[1]
        assert feed_entry.title == "-Info.terremoto: NO DATE"
        assert feed_entry.external_id == "ITEM_ID"
        self.assertIsNone(feed_entry.image_url)
        self.assertIsNone(feed_entry.region)
        self.assertIsNone(feed_entry.magnitude)
        self.assertIsNone(feed_entry.published)

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_with_category(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (40.38, -3.72)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('ign_sismologia_feed.xml')

        feed = IgnSismologiaFeed(home_coordinates,
                                 filter_minimum_magnitude=3.0)
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 2

        feed_entry = entries[0]
        assert feed_entry.title == "-Info.terremoto: 18/03/2019 19:34:55"
        assert feed_entry.external_id == "http://www.ign.es/web/ign/portal/" \
                                         "sis-catalogo-terremotos/-/" \
                                         "catalogo-terremotos/" \
                                         "detailTerremoto?evid=es2019dfcfp"

        feed_entry = entries[1]
        assert feed_entry.title == "-Info.terremoto: 18/03/2019 1:11:35"
        assert feed_entry.external_id == "http://www.ign.es/web/ign/portal/" \
                                         "sis-catalogo-terremotos/-/" \
                                         "catalogo-terremotos/" \
                                         "detailTerremoto?evid=es2019dejoe"



"""
IGN Instituto Geográfico Nacional Sismología Feed.

Fetches GeoRSS feed from ING Instituto Geográfico Nacional Sismología.
"""
from datetime import datetime
import dateparser as dateparser

from typing import Optional

from georss_client import GeoRssFeed, FeedEntry
from georss_client.consts import CUSTOM_ATTRIBUTE

ATTRIBUTION = "Instituto Geográfico Nacional"

IMAGE_URL_PATTERN = "http://www.ign.es/web/resources/sismologia/www/" \
                    "dir_images_terremotos/detalle/{}.gif"

REGEXP_ATTR_MAGNITUDE = r'magnitud (?P<{}>[^ ]+) '\
    .format(CUSTOM_ATTRIBUTE)
REGEXP_ATTR_REGION = r'magnitud [^ ]+ en (?P<{}>[A-ZÁÉÓÜÑ0-9 \-\.]+) en'\
    .format(CUSTOM_ATTRIBUTE)
REGEXP_ATTR_PUBLISHED_DATE = r'-Info.terremoto: (?P<{}>.+)$'\
    .format(CUSTOM_ATTRIBUTE)
REGEXP_ATTR_SHORT_ID = r'http:\/\/www\.ign\.es\/web\/ign\/portal\/' \
                       r'sis-catalogo-terremotos\/-\/catalogo-terremotos\/' \
                       r'detailTerremoto\?evid=(?P<{}>\w+)$'\
    .format(CUSTOM_ATTRIBUTE)

URL = "http://www.ign.es/ign/RssTools/sismologia.xml"


class IgnSismologiaFeed(GeoRssFeed):
    """IGN Instituto Geográfico Nacional Sismología feed."""

    def __init__(self, home_coordinates, filter_radius=None,
                 filter_minimum_magnitude=None):
        """Initialise this service."""
        super().__init__(home_coordinates, URL, filter_radius=filter_radius)
        self._filter_minimum_magnitude = filter_minimum_magnitude

    def __repr__(self):
        """Return string representation of this feed."""
        return '<{}(home={}, url={}, radius={}, magnitude={})>'.format(
            self.__class__.__name__, self._home_coordinates, self._url,
            self._filter_radius, self._filter_minimum_magnitude)

    def _new_entry(self, home_coordinates, rss_entry, global_data):
        """Generate a new entry."""
        return IgnSismologiaFeedEntry(home_coordinates, rss_entry)

    def _filter_entries(self, entries):
        """Filter the provided entries."""
        entries = super()._filter_entries(entries)
        if self._filter_minimum_magnitude:
            # Return only entries that have an actual magnitude value, and
            # the value is equal or above the defined threshold.
            return list(filter(lambda entry:
                               entry.magnitude and entry.magnitude >= self.
                               _filter_minimum_magnitude, entries))
        return entries


class IgnSismologiaFeedEntry(FeedEntry):
    """IGN Instituto Geográfico Nacional Sismología feed entry."""

    def __init__(self, home_coordinates, rss_entry):
        """Initialise this service."""
        super().__init__(home_coordinates, rss_entry)

    @property
    def attribution(self) -> str:
        """Return the attribution of this entry."""
        return ATTRIBUTION

    @property
    def published(self) -> Optional[datetime]:
        """Return the published date of this entry."""
        published_date = self._search_in_title(REGEXP_ATTR_PUBLISHED_DATE)
        if published_date:
            published_date = dateparser.parse(published_date)
        return published_date

    @property
    def magnitude(self) -> Optional[float]:
        """Return the magnitude of this entry."""
        magnitude = self._search_in_description(REGEXP_ATTR_MAGNITUDE)
        if magnitude:
            magnitude = float(magnitude)
        return magnitude

    @property
    def region(self) -> Optional[float]:
        """Return the region of this entry."""
        return self._search_in_description(REGEXP_ATTR_REGION)

    def _short_id(self) -> Optional[str]:
        """Return the short id of this entry."""
        return self._search_in_external_id(REGEXP_ATTR_SHORT_ID)

    @property
    def image_url(self) -> Optional[str]:
        """Return the image url of this entry."""
        short_id = self._short_id()
        if short_id:
            return IMAGE_URL_PATTERN.format(short_id)
        return None



"""Tests for XML parser."""
import datetime
import unittest

from georss_client.xml_parser import XmlParser
from georss_client.xml_parser.geometry import Point, Polygon
from tests.utils import load_fixture


class TestXmlParser(unittest.TestCase):
    """Test the XML parser."""

    def test_simple_1(self):
        """Test parsing various actual XML files."""
        xml_parser = XmlParser()
        xml = load_fixture('xml_parser_simple_1.xml')
        feed = xml_parser.parse(xml)
        self.assertIsNotNone(feed)
        self.assertIsNotNone(feed.entries)
        assert len(feed.entries) == 1

    def test_simple_2(self):
        """Test parsing various actual XML files."""
        xml_parser = XmlParser()
        xml = load_fixture('xml_parser_simple_2.xml')
        feed = xml_parser.parse(xml)
        self.assertIsNotNone(feed)
        self.assertIsNotNone(feed.entries)
        assert len(feed.entries) == 1

    def test_simple_3(self):
        """Test parsing various actual XML files."""
        xml_parser = XmlParser()
        xml = load_fixture('xml_parser_simple_3.xml')
        feed = xml_parser.parse(xml)
        self.assertIsNone(feed)

    def test_complex_1(self):
        """Test parsing various actual XML files."""
        xml_parser = XmlParser()
        xml = load_fixture('xml_parser_complex_1.xml')
        feed = xml_parser.parse(xml)
        self.assertIsNotNone(feed)

        assert feed.title == "Feed Title 1"
        assert feed.subtitle == "Feed Subtitle 1"
        assert feed.description == "Feed Description 1"
        assert feed.summary == "Feed Description 1"
        assert feed.content == "Feed Description 1"
        assert feed.link == "Feed Link 1"
        assert feed.published_date \
            == datetime.datetime(2018, 12, 9, 8, 30,
                                 tzinfo=datetime.timezone.utc)
        assert feed.pub_date \
            == datetime.datetime(2018, 12, 9, 8, 30,
                                 tzinfo=datetime.timezone.utc)
        assert feed.updated_date \
            == datetime.datetime(2018, 12, 9, 8, 45,
                                 tzinfo=datetime.timezone.utc)
        assert feed.last_build_date \
            == datetime.datetime(2018, 12, 9, 8, 45,
                                 tzinfo=datetime.timezone.utc)
        assert feed.copyright == "Feed Copyright 1"
        assert feed.rights == "Feed Copyright 1"
        assert feed.generator == "Feed Generator 1"
        assert feed.language == "Feed Language 1"
        assert feed.docs == "http://docs.url/documentation.html"
        assert feed.ttl == 42
        assert feed.author == "Feed Author 1"
        assert feed.contributor == "Feed Author 1"
        assert feed.managing_editor == "Feed Author 1"
        assert feed.category == ["Feed Category 1"]
        self.assertIsNotNone(feed.image)
        assert feed.image.title == "Image Title 1"
        assert feed.image.url == "http://image.url/image.png"
        assert feed.image.link == "http://feed.link/feed.rss"
        assert feed.image.description == "Image Description 1"
        assert feed.image.width == 123
        assert feed.image.height == 234
        assert feed.get_additional_attribute('random') == "Feed Random 1"
        assert repr(feed) == "<Feed(Feed Link 1)>"

        self.assertIsNotNone(feed.entries)
        assert len(feed.entries) == 6

        feed_entry = feed.entries[0]
        assert feed_entry.title == "Title 1"
        assert feed_entry.description == "Description 1"
        assert feed_entry.link == "Link 1"
        assert feed_entry.published_date \
            == datetime.datetime(2018, 12, 9, 7, 30,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.updated_date \
            == datetime.datetime(2018, 12, 9, 7, 45,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.guid == "GUID 1"
        assert feed_entry.id == "GUID 1"
        assert feed_entry.source == "Source 1"
        assert feed_entry.category == ["Category 1"]
        self.assertIsInstance(feed_entry.geometry, Point)
        assert feed_entry.geometry.latitude == -37.4567
        assert feed_entry.geometry.longitude == 149.3456
        assert feed_entry.get_additional_attribute('random') == "Random 1"
        assert repr(feed_entry) == "<FeedItem(GUID 1)>"

        feed_entry = feed.entries[1]
        assert feed_entry.title == "Title 2"
        assert feed_entry.description == "Description 2"
        assert feed_entry.link == "Link 2"
        assert feed_entry.published_date \
            == datetime.datetime(2018, 12, 9, 7, 35,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.updated_date \
            == datetime.datetime(2018, 12, 9, 7, 50,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.guid == "GUID 2"
        assert feed_entry.category == ["Category 2"]
        self.assertIsInstance(feed_entry.geometry, Point)
        assert feed_entry.geometry.latitude == -37.5678
        assert feed_entry.geometry.longitude == 149.4567

        feed_entry = feed.entries[2]
        assert feed_entry.title == "Title 3"
        assert feed_entry.description == "Description 3"
        assert feed_entry.published_date \
            == datetime.datetime(2018, 12, 9, 7, 40,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.updated_date \
            == datetime.datetime(2018, 12, 9, 7, 55,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.guid == "GUID 3"
        assert feed_entry.category == ["Category 3A",
                                       "Category 3B",
                                       "Category 3C"]
        self.assertIsInstance(feed_entry.geometry, Point)
        assert feed_entry.geometry.latitude == -37.6789
        assert feed_entry.geometry.longitude == 149.5678

        feed_entry = feed.entries[3]
        assert feed_entry.title == "Title 4"
        assert feed_entry.description == "Description 4"
        assert feed_entry.author == "Author 4"
        assert feed_entry.contributor == "Author 4"
        assert feed_entry.category == ["Category 4A",
                                       "Category 4B"]
        assert feed_entry.published_date == datetime.datetime(
            2018, 9, 30, 21, 36, 48,
            tzinfo=datetime.timezone(datetime.timedelta(hours=10), 'AEST'))
        self.assertIsInstance(feed_entry.geometry, Point)
        assert feed_entry.geometry.latitude == -37.789
        assert feed_entry.geometry.longitude == 149.6789

        feed_entry = feed.entries[4]
        assert feed_entry.title == "Title 5"
        assert feed_entry.description == "Description 5"
        assert feed_entry.published_date == datetime.datetime(
            2018, 9, 20, 18, 1, 55,
            tzinfo=datetime.timezone(datetime.timedelta(hours=2), 'CEST'))
        self.assertIsInstance(feed_entry.geometry, Polygon)
        assert feed_entry.geometry.centroid.latitude == -30.32
        assert feed_entry.geometry.centroid.longitude == 150.32

        feed_entry = feed.entries[5]
        assert feed_entry.title == "Title 6"
        assert feed_entry.description == "Description 6"
        assert feed_entry.published_date == datetime.datetime(
            2018, 10, 7, 19,52,
            tzinfo=datetime.timezone(datetime.timedelta(hours=-7), 'PDT'))
        self.assertIsInstance(feed_entry.geometry, Polygon)
        assert feed_entry.geometry.centroid.latitude == -30.32
        assert feed_entry.geometry.centroid.longitude == 150.32

    def test_complex_2(self):
        """Test parsing various actual XML files."""
        xml_parser = XmlParser()
        xml = load_fixture('xml_parser_complex_2.xml')
        feed = xml_parser.parse(xml)
        self.assertIsNotNone(feed)

        assert feed.title == "Feed Title 1"
        assert feed.subtitle == "Feed Subtitle 1"
        assert feed.ttl == "INVALID"
        assert feed.author == "Author 1"
        assert feed.last_build_date \
            == datetime.datetime(2018, 12, 9, 9, 0,
                                 tzinfo=datetime.timezone.utc)
        assert feed.updated_date \
            == datetime.datetime(2018, 12, 9, 9, 0,
                                 tzinfo=datetime.timezone.utc)
        assert feed.copyright == "Feed Rights 1"
        assert feed.rights == "Feed Rights 1"
        assert feed.generator == "Feed Generator 1"
        self.assertIsNotNone(feed.image)
        assert feed.image.title == "Image Title 1"
        assert feed.image.url == "http://image.url/image.png"
        assert feed.image.link == "http://feed.link/feed.rss"
        self.assertIsNone(feed.image.description)
        self.assertIsNone(feed.image.width)
        self.assertIsNone(feed.image.height)
        self.assertIsNone(feed.docs)

        self.assertIsNotNone(feed.entries)
        assert len(feed.entries) == 1

        feed_entry = feed.entries[0]
        assert feed_entry.title == "Title 6"
        self.assertIsNone(feed_entry.published_date)

    def test_complex_3(self):
        """Test parsing various actual XML files."""
        xml_parser = XmlParser()
        xml = load_fixture('xml_parser_complex_3.xml')
        feed = xml_parser.parse(xml)
        self.assertIsNotNone(feed)

        self.assertIsNone(feed.title)
        self.assertIsNone(feed.subtitle)
        self.assertIsNone(feed.description)
        self.assertIsNone(feed.language)
        self.assertIsNone(feed.published_date)
        self.assertIsNone(feed.last_build_date)
        self.assertIsNone(feed.ttl)
        assert feed.rights == "Feed Rights 1"
        self.assertIsNone(feed.image)

        self.assertIsNotNone(feed.entries)
        assert len(feed.entries) == 2

        feed_entry = feed.entries[0]
        self.assertIsNone(feed_entry.title)
        self.assertIsNone(feed_entry.published_date)
        self.assertIsNone(feed_entry.geometry)

        feed_entry = feed.entries[1]
        self.assertIsNone(feed_entry.title)
        self.assertIsNone(feed_entry.geometry)


class TestGeometries(unittest.TestCase):
    """Test geometries."""

    def test_point(self):
        """Test point."""
        point = Point(-37.1234, 149.2345)
        assert point.latitude == -37.1234
        assert point.longitude == 149.2345
        assert repr(point) == "<Point(latitude=-37.1234, longitude=149.2345)>"

    def test_polygon(self):
        """Test polygon."""
        polygon = Polygon([
            Point(-30.1, 150.1),
            Point(-30.2, 150.2),
            Point(-30.4, 150.4),
            Point(-30.8, 150.8),
            Point(-30.1, 150.1)
        ])
        assert len(polygon.points) == 5
        assert polygon.centroid.latitude == -30.32
        assert polygon.centroid.longitude == 150.32
        assert repr(polygon) == "<Polygon(centroid=" \
                                "<Point(latitude=-30.32, longitude=150.32)>)>"



"""Test for the generic georss feed."""
import datetime
import requests
import unittest
from unittest import mock

from georss_client import UPDATE_ERROR, UPDATE_OK
from georss_client.generic_feed import GenericFeed
from tests.utils import load_fixture


class TestGenericFeed(unittest.TestCase):
    """Test the generic feed."""

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (-31.0, 151.0)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = load_fixture('generic_feed_1.xml')

        feed = GenericFeed(home_coordinates, None)
        assert repr(feed) == "<GenericFeed(home=(-31.0, 151.0), url=None, " \
                             "radius=None, categories=None)>"
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 5

        feed_entry = entries[0]
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "1234"
        assert feed_entry.category == "Category 1"
        assert feed_entry.published == datetime.datetime(2018, 9, 23, 8, 30)
        assert feed_entry.updated == datetime.datetime(2018, 9, 23, 8, 35)
        assert feed_entry.coordinates == (-37.2345, 149.1234)
        self.assertAlmostEqual(feed_entry.distance_to_home, 714.4, 1)

        feed_entry = entries[1]
        assert feed_entry.title == "Title 2"
        assert feed_entry.external_id == "2345"
        assert feed_entry.attribution == "Attribution 1"
        assert repr(feed_entry) == "<GenericFeedEntry(id=2345)>"

        feed_entry = entries[2]
        assert feed_entry.title == "Title 3"
        assert feed_entry.external_id == "Title 3"

        feed_entry = entries[3]
        self.assertIsNone(feed_entry.title)
        assert feed_entry.external_id == -7266545992534134585

        feed_entry = entries[4]
        assert feed_entry.title == "Title 5"
        assert feed_entry.external_id == "5678"

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_feed_2(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (-31.0, 151.0)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = load_fixture('generic_feed_2.xml')

        feed = GenericFeed(home_coordinates, None)
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 1

        feed_entry = entries[0]
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "1234"
        assert feed_entry.category == "Category 1"
        assert feed_entry.coordinates == (-37.2345, 149.1234)
        self.assertAlmostEqual(feed_entry.distance_to_home, 714.4, 1)

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_feed_3(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (-31.0, 151.0)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = load_fixture('generic_feed_3.xml')

        feed = GenericFeed(home_coordinates, None)
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 3

        feed_entry = entries[0]
        assert feed_entry.external_id == "1234"
        assert feed_entry.coordinates == (-34.93728111547821,
                                          148.59710883878262)
        self.assertAlmostEqual(feed_entry.distance_to_home, 491.7, 1)

        feed_entry = entries[1]
        assert feed_entry.external_id == "2345"
        assert feed_entry.coordinates == (-34.937170989, 148.597182317)
        self.assertAlmostEqual(feed_entry.distance_to_home, 491.8, 1)

        feed_entry = entries[2]
        assert feed_entry.external_id == "3456"
        assert feed_entry.coordinates == (-29.962746645660683,
                                          152.43090880416074)
        self.assertAlmostEqual(feed_entry.distance_to_home, 176.5, 1)

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_with_radius_filtering(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (-37.0, 150.0)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = load_fixture('generic_feed_1.xml')

        feed = GenericFeed(home_coordinates, None, filter_radius=90.0)
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 4
        self.assertAlmostEqual(entries[0].distance_to_home, 82.0, 1)
        self.assertAlmostEqual(entries[1].distance_to_home, 77.0, 1)
        self.assertAlmostEqual(entries[2].distance_to_home, 84.6, 1)

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_with_radius_and_category_filtering(self, mock_session,
                                                          mock_request):
        """Test updating feed is ok."""
        home_coordinates = (-37.0, 150.0)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = load_fixture('generic_feed_1.xml')

        feed = GenericFeed(home_coordinates, None, filter_radius=90.0,
                           filter_categories=['Category 2'])
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 1
        self.assertAlmostEqual(entries[0].distance_to_home, 77.0, 1)

        feed = GenericFeed(home_coordinates, None, filter_radius=90.0,
                           filter_categories=['Category 4'])
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 0

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_error(self, mock_session, mock_request):
        """Test updating feed results in error."""
        home_coordinates = (-31.0, 151.0)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = False

        feed = GenericFeed(home_coordinates, None)
        status, entries = feed.update()
        assert status == UPDATE_ERROR

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_with_request_exception(self, mock_session, mock_request):
        """Test updating feed raises exception."""
        home_coordinates = (-31.0, 151.0)
        mock_session.return_value.__enter__.return_value.send\
            .side_effect = requests.exceptions.RequestException

        feed = GenericFeed(home_coordinates, None)
        status, entries = feed.update()
        assert status == UPDATE_ERROR
        self.assertIsNone(entries)



"""Tests for base classes."""
import unittest
from unittest.mock import MagicMock

from georss_client import GeoRssDistanceHelper, FeedEntry
from georss_client.xml_parser.geometry import Point, Polygon


class TestGeoRssDistanceHelper(unittest.TestCase):
    """Tests for the GeoJSON distance helper."""

    def test_extract_coordinates_from_point(self):
        """Test extracting coordinates from point."""
        mock_point = Point(-30.0, 151.0)
        latitude, longitude = GeoRssDistanceHelper.\
            extract_coordinates(mock_point)
        assert latitude == -30.0
        assert longitude == 151.0

    def test_extract_coordinates_from_polygon(self):
        """Test extracting coordinates from polygon."""
        mock_polygon = Polygon([Point(-30.0, 151.0),
                                Point(-30.0, 151.5),
                                Point(-30.5, 151.5),
                                Point(-30.5, 151.0),
                                Point(-30.0, 151.0)])
        latitude, longitude = GeoRssDistanceHelper.\
            extract_coordinates(mock_polygon)
        self.assertAlmostEqual(latitude, -30.2, 1)
        self.assertAlmostEqual(longitude, 151.2, 1)

    def test_extract_coordinates_from_unsupported_geometry(self):
        """Test extracting coordinates from unsupported geometry."""
        mock_unsupported_geometry = MagicMock()
        latitude, longitude = GeoRssDistanceHelper.\
            extract_coordinates(mock_unsupported_geometry)
        self.assertIsNone(latitude)
        self.assertIsNone(longitude)

    def test_distance_to_point(self):
        """Test calculating distance to point."""
        home_coordinates = [-31.0, 150.0]
        mock_point = Point(-30.0, 151.0)
        distance = GeoRssDistanceHelper.\
            distance_to_geometry(home_coordinates, mock_point)
        self.assertAlmostEqual(distance, 146.8, 1)

    def test_distance_to_polygon(self):
        """Test calculating distance to point."""
        home_coordinates = [-31.0, 150.0]
        mock_polygon = Polygon([Point(-30.0, 151.0),
                                Point(-30.0, 151.5),
                                Point(-30.5, 151.5),
                                Point(-30.5, 151.0),
                                Point(-30.0, 151.0)])
        distance = GeoRssDistanceHelper.\
            distance_to_geometry(home_coordinates, mock_polygon)
        self.assertAlmostEqual(distance, 110.6, 1)

    def test_distance_to_unsupported_geometry(self):
        """Test calculating distance to unsupported geometry."""
        home_coordinates = [-31.0, 150.0]
        mock_unsupported_geometry = MagicMock()
        distance = GeoRssDistanceHelper.\
            distance_to_geometry(home_coordinates, mock_unsupported_geometry)
        assert distance == float("inf")


class TestFeedEntry(unittest.TestCase):

    def test_feed_entry(self):
        """Test feed entry behaviour."""
        feed_entry = FeedEntry(None, None)
        assert repr(feed_entry) == "<FeedEntry(id=None)>"
        self.assertIsNone(feed_entry.geometry)
        self.assertIsNone(feed_entry.coordinates)
        self.assertIsNone(feed_entry.title)
        self.assertIsNone(feed_entry.category)
        self.assertIsNone(feed_entry.attribution)
        self.assertIsNone(feed_entry.description)

        mock_rss_entry = dict()
        feed_entry = FeedEntry(None, mock_rss_entry)
        assert repr(feed_entry) == "<FeedEntry(id=None)>"



"""Test for the QFES Bushfire Alert feed."""
import datetime
import unittest
from unittest import mock

from georss_client import UPDATE_OK
from georss_client.qfes_bushfire_alert_feed import \
    QfesBushfireAlertFeed
from tests.utils import load_fixture


class TestQfesBushfireAlertFeed(unittest.TestCase):
    """Test the QFES Bushfire Alert feed."""

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (-31.0, 151.0)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('qfes_bushfire_alert_feed.xml')

        feed = QfesBushfireAlertFeed(home_coordinates)
        assert repr(feed) == "<QfesBushfireAlertFeed(home=(-31.0, 151.0), " \
                             "url=https://www.qfes.qld.gov.au/data/alerts/" \
                             "bushfireAlert.xml, radius=None, categories=" \
                             "None)>"
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 2

        feed_entry = entries[0]
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "1234"
        assert feed_entry.coordinates == (-32.2345, 149.1234)
        self.assertAlmostEqual(feed_entry.distance_to_home, 224.5, 1)
        assert feed_entry.published == datetime.datetime(2018, 9, 27, 8, 0)
        assert feed_entry.updated == datetime.datetime(2018, 9, 27, 8, 30)
        assert feed_entry.status == "Status 1"
        assert feed_entry.attribution == "Author 1"
        assert repr(feed_entry) == "<QfesBushfireAlertFeedEntry(id=1234)>"

        feed_entry = entries[1]
        assert feed_entry.title == "Title 2"
        self.assertIsNone(feed_entry.published)
        self.assertIsNone(feed_entry.updated)

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_with_category(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (-31.0, 151.0)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('qfes_bushfire_alert_feed.xml')

        feed = QfesBushfireAlertFeed(
            home_coordinates,
            filter_categories=['Category 1'])
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 1

        feed_entry = entries[0]
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "1234"



"""
Base class for GeoRSS services.

Fetches GeoRSS feed from URL to be defined by sub-class.
"""
from datetime import datetime
import logging
import re

import requests
from haversine import haversine
from typing import Optional

from georss_client.consts import ATTR_ATTRIBUTION, CUSTOM_ATTRIBUTE
from georss_client.xml_parser import XmlParser
from georss_client.xml_parser.geometry import Point, Polygon

_LOGGER = logging.getLogger(__name__)

UPDATE_OK = 'OK'
UPDATE_OK_NO_DATA = 'OK_NO_DATA'
UPDATE_ERROR = 'ERROR'


class GeoRssFeed:
    """GeoRSS feed base class."""

    def __init__(self, home_coordinates, url, filter_radius=None,
                 filter_categories=None):
        """Initialise this service."""
        self._home_coordinates = home_coordinates
        self._filter_radius = filter_radius
        self._filter_categories = filter_categories
        self._url = url
        self._request = requests.Request(method="GET", url=url).prepare()

    def __repr__(self):
        """Return string representation of this feed."""
        return '<{}(home={}, url={}, radius={}, categories={})>'.format(
            self.__class__.__name__, self._home_coordinates, self._url,
            self._filter_radius, self._filter_categories)

    def _new_entry(self, home_coordinates, rss_entry, global_data):
        """Generate a new entry."""
        pass

    def _additional_namespaces(self):
        """Provide additional namespaces, relevant for this feed."""
        pass

    def update(self):
        """Update from external source and return filtered entries."""
        status, data = self._fetch()
        if status == UPDATE_OK:
            if data:
                entries = []
                global_data = self._extract_from_feed(data)
                # Extract data from feed entries.
                for rss_entry in data.entries:
                    entries.append(self._new_entry(self._home_coordinates,
                                                   rss_entry, global_data))
                return UPDATE_OK, self._filter_entries(entries)
            else:
                # Should not happen.
                return UPDATE_OK, None
        elif status == UPDATE_OK_NO_DATA:
            # Happens for example if the server returns 304
            return UPDATE_OK_NO_DATA, None
        else:
            # Error happened while fetching the feed.
            return UPDATE_ERROR, None

    def _fetch(self):
        """Fetch GeoRSS data from external source."""
        try:
            with requests.Session() as session:
                response = session.send(self._request, timeout=10)
            if response.ok:
                parser = XmlParser(self._additional_namespaces())
                feed_data = parser.parse(response.text)
                self.parser = parser
                self.feed_data = feed_data
                return UPDATE_OK, feed_data
            else:
                _LOGGER.warning(
                    "Fetching data from %s failed with status %s",
                    self._request.url, response.status_code)
                return UPDATE_ERROR, None
        except requests.exceptions.RequestException as request_ex:
            _LOGGER.warning("Fetching data from %s failed with %s",
                            self._request.url, request_ex)
            return UPDATE_ERROR, None

    def _filter_entries(self, entries):
        """Filter the provided entries."""
        filtered_entries = entries
        _LOGGER.debug("Entries before filtering %s", filtered_entries)
        # Always remove entries without geometry
        filtered_entries = list(
            filter(lambda entry:
                   entry.geometry is not None,
                   filtered_entries))
        # Filter by distance.
        if self._filter_radius:
            filtered_entries = list(
                filter(lambda entry:
                       entry.distance_to_home <= self._filter_radius,
                       filtered_entries))
        # Filter by category.
        if self._filter_categories:
            filtered_entries = list(
                filter(lambda entry:
                       len({entry.category}.intersection(
                           self._filter_categories)) > 0,
                       filtered_entries))
        _LOGGER.debug("Entries after filtering %s", filtered_entries)
        return filtered_entries

    def _extract_from_feed(self, feed):
        """Extract global metadata from feed."""
        global_data = {}
        author = feed.author
        if author:
            global_data[ATTR_ATTRIBUTION] = author
        return global_data


class FeedEntry:
    """Feed entry base class."""

    def __init__(self, home_coordinates, rss_entry):
        """Initialise this feed entry."""
        self._home_coordinates = home_coordinates
        self._rss_entry = rss_entry

    def __repr__(self):
        """Return string representation of this entry."""
        return '<{}(id={})>'.format(self.__class__.__name__, self.external_id)

    @property
    def geometry(self):
        """Return all geometry details of this entry."""
        if self._rss_entry:
            return self._rss_entry.geometry
        return None

    @property
    def coordinates(self):
        """Return the best coordinates (latitude, longitude) of this entry."""
        if self.geometry:
            return GeoRssDistanceHelper.extract_coordinates(self.geometry)
        return None

    @property
    def external_id(self) -> Optional[str]:
        """Return the external id of this entry."""
        if self._rss_entry:
            external_id = self._rss_entry.guid
            if not external_id:
                external_id = self.title
            if not external_id:
                # Use geometry as ID as a fallback.
                external_id = hash(self.coordinates)
            return external_id
        return None

    def _search_in_external_id(self, regexp):
        """Find a sub-string in the entry's external id."""
        if self.external_id:
            match = re.search(regexp, self.external_id)
            if match:
                return match.group(CUSTOM_ATTRIBUTE)
        return None

    @property
    def title(self) -> Optional[str]:
        """Return the title of this entry."""
        if self._rss_entry:
            return self._rss_entry.title
        return None

    def _search_in_title(self, regexp):
        """Find a sub-string in the entry's title."""
        if self.title:
            match = re.search(regexp, self.title)
            if match:
                return match.group(CUSTOM_ATTRIBUTE)
        return None

    @property
    def category(self) -> Optional[str]:
        """Return the category of this entry."""
        if self._rss_entry and self._rss_entry.category \
                and isinstance(self._rss_entry.category, list):
            # To keep this simple, just return the first category.
            return self._rss_entry.category[0]
        return None

    @property
    def attribution(self) -> Optional[str]:
        """Return the attribution of this entry."""
        return None

    @property
    def distance_to_home(self):
        """Return the distance in km of this entry to the home coordinates."""
        return GeoRssDistanceHelper.distance_to_geometry(
            self._home_coordinates, self.geometry)

    @property
    def description(self) -> Optional[str]:
        """Return the description of this entry."""
        if self._rss_entry and self._rss_entry.description:
            return self._rss_entry.description
        return None

    @property
    def published(self) -> Optional[datetime]:
        """Return the published date of this entry."""
        if self._rss_entry:
            return self._rss_entry.published_date
        return None

    @property
    def updated(self) -> Optional[datetime]:
        """Return the updated date of this entry."""
        if self._rss_entry:
            return self._rss_entry.updated_date
        return None

    def _search_in_description(self, regexp):
        """Find a sub-string in the entry's description."""
        if self.description:
            match = re.search(regexp, self.description)
            if match:
                return match.group(CUSTOM_ATTRIBUTE)
        return None


class GeoRssDistanceHelper:
    """Helper to calculate distances between GeoRSS geometries."""

    def __init__(self):
        """Initialize the geo distance helper."""
        pass

    @staticmethod
    def extract_coordinates(geometry):
        """Extract the best coordinates from the feature for display."""
        latitude = longitude = None
        if isinstance(geometry, Point):
            # Just extract latitude and longitude directly.
            latitude, longitude = geometry.latitude, geometry.longitude
        elif isinstance(geometry, Polygon):
            centroid = geometry.centroid
            latitude, longitude = centroid.latitude, centroid.longitude
            _LOGGER.debug("Centroid of %s is %s", geometry,
                          (latitude, longitude))
        else:
            _LOGGER.debug("Not implemented: %s", type(geometry))
        return latitude, longitude

    @staticmethod
    def distance_to_geometry(home_coordinates, geometry):
        """Calculate the distance between home coordinates and geometry."""
        distance = float("inf")
        if isinstance(geometry, Point):
            distance = GeoRssDistanceHelper._distance_to_point(
                home_coordinates, geometry)
        elif isinstance(geometry, Polygon):
            distance = GeoRssDistanceHelper._distance_to_polygon(
                home_coordinates, geometry)
        else:
            _LOGGER.debug("Not implemented: %s", type(geometry))
        return distance

    @staticmethod
    def _distance_to_point(home_coordinates, point):
        """Calculate the distance between home coordinates and the point."""
        # Swap coordinates to match: (latitude, longitude).
        return GeoRssDistanceHelper._distance_to_coordinates(
            home_coordinates, (point.latitude, point.longitude))

    @staticmethod
    def _distance_to_polygon(home_coordinates, polygon):
        """Calculate the distance between home coordinates and the polygon."""
        distance = float("inf")
        # Calculate distance from polygon by calculating the distance
        # to each point of the polygon but not to each edge of the
        # polygon; should be good enough
        for point in polygon.points:
            distance = min(distance,
                           GeoRssDistanceHelper._distance_to_coordinates(
                               home_coordinates,
                               (point.latitude, point.longitude)))
        return distance

    @staticmethod
    def _distance_to_coordinates(home_coordinates, coordinates):
        """Calculate the distance between home coordinates and the
        coordinates."""
        # Expecting coordinates in format: (latitude, longitude).
        return haversine(coordinates, home_coordinates)



"""
Geometry models.
"""
from typing import Optional


class Geometry:
    """Represents a geometry."""


class Point(Geometry):
    """Represents a point."""

    def __init__(self, latitude, longitude):
        """Initialise point."""
        self._latitude = latitude
        self._longitude = longitude

    def __repr__(self):
        """Return string representation of this point."""
        return '<{}(latitude={}, longitude={})>'.format(
            self.__class__.__name__, self.latitude, self.longitude)

    @property
    def latitude(self) -> Optional[float]:
        """Return the latitude of this point."""
        return self._latitude

    @property
    def longitude(self) -> Optional[float]:
        """Return the longitude of this point."""
        return self._longitude


class Polygon(Geometry):
    """Represents a polygon."""

    def __init__(self, points):
        """Initialise polygon."""
        self._points = points

    def __repr__(self):
        """Return string representation of this polygon."""
        return '<{}(centroid={})>'.format(
            self.__class__.__name__, self.centroid)

    @property
    def points(self) -> Optional[list]:
        """Return the points of this polygon."""
        return self._points

    @property
    def centroid(self) -> Point:
        """Find the polygon's centroid as a best approximation."""
        longitudes_list = [point.longitude for point in self.points]
        latitudes_list = [point.latitude for point in self.points]
        number_of_points = len(self.points)
        longitude = sum(longitudes_list) / number_of_points
        latitude = sum(latitudes_list) / number_of_points
        return Point(latitude, longitude)



"""
INGV Centro Nazionale Terremoti (Earthquakes) Feed.

Fetches GeoRSS feed from INGV Centro Nazionale Terremoti.
"""
from datetime import datetime

from typing import Optional

from georss_client import GeoRssFeed, FeedEntry
from georss_client.consts import CUSTOM_ATTRIBUTE, ATTR_ATTRIBUTION

REGEXP_ATTR_MAGNITUDE = r'Magnitude\(M.{{0,3}}\) (?P<{}>[^ ]+) '\
    .format(CUSTOM_ATTRIBUTE)
REGEXP_ATTR_REGION = r'Magnitude\(M.{{0,3}}\) [^ ]+[ ]+-[ ]+(?P<{}>.+)$'\
    .format(CUSTOM_ATTRIBUTE)

URL = "http://cnt.rm.ingv.it/feed/atom/all_week"


class IngvCentroNazionaleTerremotiFeed(GeoRssFeed):
    """INGV Centro Nazionale Terremoti feed."""

    def __init__(self, home_coordinates, filter_radius=None,
                 filter_minimum_magnitude=None):
        """Initialise this service."""
        super().__init__(home_coordinates, URL, filter_radius=filter_radius)
        self._filter_minimum_magnitude = filter_minimum_magnitude

    def __repr__(self):
        """Return string representation of this feed."""
        return '<{}(home={}, url={}, radius={}, magnitude={})>'.format(
            self.__class__.__name__, self._home_coordinates, self._url,
            self._filter_radius, self._filter_minimum_magnitude)

    def _new_entry(self, home_coordinates, rss_entry, global_data):
        """Generate a new entry."""
        attribution = None if not global_data and ATTR_ATTRIBUTION not in \
            global_data else global_data[ATTR_ATTRIBUTION]
        return IngvCentroNazionaleTerremotiFeedEntry(home_coordinates,
                                                     rss_entry, attribution)

    def _filter_entries(self, entries):
        """Filter the provided entries."""
        entries = super()._filter_entries(entries)
        if self._filter_minimum_magnitude:
            # Return only entries that have an actual magnitude value, and
            # the value is equal or above the defined threshold.
            return list(filter(lambda entry:
                               entry.magnitude and entry.magnitude >= self.
                               _filter_minimum_magnitude, entries))
        return entries


class IngvCentroNazionaleTerremotiFeedEntry(FeedEntry):
    """INGV Centro Nazionale Terremoti feed entry."""

    def __init__(self, home_coordinates, rss_entry, attribution):
        """Initialise this service."""
        super().__init__(home_coordinates, rss_entry)
        self._attribution = attribution

    @property
    def attribution(self) -> str:
        """Return the attribution of this entry."""
        return self._attribution

    @property
    def magnitude(self) -> Optional[float]:
        """Return the magnitude of this entry."""
        magnitude = self._search_in_title(REGEXP_ATTR_MAGNITUDE)
        if magnitude:
            magnitude = float(magnitude)
        return magnitude

    @property
    def region(self) -> Optional[float]:
        """Return the region of this entry."""
        return self._search_in_title(REGEXP_ATTR_REGION)



"""Test for the Natural Resources Canada Earthquakes feed."""
import datetime
import unittest
from unittest import mock

from georss_client import UPDATE_OK
from georss_client.exceptions import GeoRssException
from georss_client.natural_resources_canada_earthquakes_feed import \
    NaturalResourcesCanadaEarthquakesFeed
from tests.utils import load_fixture


class TestNaturalResourcesCanadaEarthquakesFeed(unittest.TestCase):
    """Test the Natural Resources Canada Earthquakes feed."""

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_en(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (49.25, -123.1)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('natural_resources_canada_earthquakes_en_feed.xml')

        feed = NaturalResourcesCanadaEarthquakesFeed(home_coordinates, 'en')
        assert repr(feed) == "<NaturalResourcesCanadaEarthquakesFeed(" \
                             "home=(49.25, -123.1), url=http://www." \
                             "earthquakescanada.nrcan.gc.ca/index-en.php?" \
                             "tpl_region=canada&tpl_output=rss, " \
                             "radius=None, magnitude=None)>"
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 2

        feed_entry = entries[0]
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "1234"
        assert feed_entry.coordinates == (44.11, -66.23)
        self.assertAlmostEqual(feed_entry.distance_to_home, 4272.4, 1)
        assert feed_entry.published \
            == datetime.datetime(2018, 9, 29, 8, 30,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.category == "Category 1"
        assert feed_entry.magnitude == 4.5
        assert feed_entry.attribution == "Natural Resources Canada"
        assert repr(feed_entry) == "<NaturalResourcesCanadaEarthquakes" \
                                   "FeedEntry(id=1234)>"

        feed_entry = entries[1]
        assert feed_entry.title == "Title 2"
        self.assertIsNone(feed_entry.published)

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_en_with_magnitude(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (49.25, -123.1)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('natural_resources_canada_earthquakes_en_feed.xml')

        feed = NaturalResourcesCanadaEarthquakesFeed(
            home_coordinates, 'en', filter_minimum_magnitude=4.0)
        assert repr(feed) == "<NaturalResourcesCanadaEarthquakesFeed(" \
                             "home=(49.25, -123.1), url=http://www." \
                             "earthquakescanada.nrcan.gc.ca/index-en.php?" \
                             "tpl_region=canada&tpl_output=rss, " \
                             "radius=None, magnitude=4.0)>"
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 1

        feed_entry = entries[0]
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "1234"

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_fr(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (49.25, -123.1)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('natural_resources_canada_earthquakes_fr_feed.xml')

        feed = NaturalResourcesCanadaEarthquakesFeed(home_coordinates, 'fr')
        assert repr(feed) == "<NaturalResourcesCanadaEarthquakesFeed(" \
                             "home=(49.25, -123.1), url=http://www." \
                             "earthquakescanada.nrcan.gc.ca/index-fr.php?" \
                             "tpl_region=canada&tpl_output=rss, " \
                             "radius=None, magnitude=None)>"
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 2

        feed_entry = entries[0]
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "1234"
        assert feed_entry.coordinates == (44.11, -66.23)
        self.assertAlmostEqual(feed_entry.distance_to_home, 4272.4, 1)
        assert feed_entry.published \
            == datetime.datetime(2018, 9, 29, 8, 30,
                                 tzinfo=datetime.timezone.utc)
        assert feed_entry.category == "Category 1"
        assert feed_entry.magnitude == 4.5
        assert feed_entry.attribution == "Ressources naturelles Canada"
        assert repr(feed_entry) == "<NaturalResourcesCanadaEarthquakes" \
                                   "FeedEntry(id=1234)>"

        feed_entry = entries[1]
        assert feed_entry.title == "Title 2"
        self.assertIsNone(feed_entry.published)

    @mock.patch("requests.Request")
    @mock.patch("requests.Session")
    def test_update_ok_fr_with_magnitude(self, mock_session, mock_request):
        """Test updating feed is ok."""
        home_coordinates = (49.25, -123.1)
        mock_session.return_value.__enter__.return_value.send\
            .return_value.ok = True
        mock_session.return_value.__enter__.return_value.send\
            .return_value.text = \
            load_fixture('natural_resources_canada_earthquakes_fr_feed.xml')

        feed = NaturalResourcesCanadaEarthquakesFeed(
            home_coordinates, 'fr', filter_minimum_magnitude=4.0)
        assert repr(feed) == "<NaturalResourcesCanadaEarthquakesFeed(" \
                             "home=(49.25, -123.1), url=http://www." \
                             "earthquakescanada.nrcan.gc.ca/index-fr.php?" \
                             "tpl_region=canada&tpl_output=rss, " \
                             "radius=None, magnitude=4.0)>"
        status, entries = feed.update()
        assert status == UPDATE_OK
        self.assertIsNotNone(entries)
        assert len(entries) == 1

        feed_entry = entries[0]
        assert feed_entry.title == "Title 1"
        assert feed_entry.external_id == "1234"

    def test_update_wrong_language(self):
        """Test invalid feed language."""
        home_coordinates = (49.25, -123.1)

        with self.assertRaises(GeoRssException):
            NaturalResourcesCanadaEarthquakesFeed(home_coordinates,
                                                  'DOES NOT EXIST')



"""
Constants.

Constants for feeds and feed entries.
"""

ATTR_ATTRIBUTION = 'attribution'
CUSTOM_ATTRIBUTE = 'custom_attribute'

XML_ATTR_HREF = '@href'
XML_ATTR_TERM = '@term'

XML_CDATA = '#text'

XML_TAG_AUTHOR = 'author'
XML_TAG_CATEGORY = 'category'
XML_TAG_CHANNEL = 'channel'
XML_TAG_CONTENT = 'content'
XML_TAG_CONTRIBUTOR = 'contributor'
XML_TAG_COPYRIGHT = 'copyright'
XML_TAG_DC_DATE = 'dc:date'
XML_TAG_DESCRIPTION = 'description'
XML_TAG_DOCS = 'docs'
XML_TAG_ENTRY = 'entry'
XML_TAG_FEED = 'feed'
XML_TAG_GENERATOR = 'generator'
XML_TAG_GEO_LAT = 'geo:lat'
XML_TAG_GEO_LONG = 'geo:long'
XML_TAG_GEO_POINT = 'geo:Point'
XML_TAG_GEORSS_POINT = 'georss:point'
XML_TAG_GEORSS_POLYGON = 'georss:polygon'
XML_TAG_GEORSS_WHERE = 'georss:where'
XML_TAG_GML_EXTERIOR = 'gml:exterior'
XML_TAG_GML_LINEAR_RING = 'gml:LinearRing'
XML_TAG_GML_POINT = 'gml:Point'
XML_TAG_GML_POLYGON = 'gml:Polygon'
XML_TAG_GML_POS = 'gml:pos'
XML_TAG_GML_POS_LIST = 'gml:posList'
XML_TAG_GUID = 'guid'
XML_TAG_HEIGHT = 'height'
XML_TAG_ID = 'id'
XML_TAG_IMAGE = 'image'
XML_TAG_ITEM = 'item'
XML_TAG_LANGUAGE = 'language'
XML_TAG_LAST_BUILD_DATE = 'lastBuildDate'
XML_TAG_LINK = 'link'
XML_TAG_MANAGING_EDITOR = 'managingEditor'
XML_TAG_NAME = 'name'
XML_TAG_PUB_DATE = 'pubDate'
XML_TAG_PUBLISHED = 'published'
XML_TAG_RIGHTS = 'rights'
XML_TAG_RSS = 'rss'
XML_TAG_SOURCE = 'source'
XML_TAG_SUBTITLE = 'subtitle'
XML_TAG_SUMMARY = 'summary'
XML_TAG_TITLE = 'title'
XML_TAG_TTL = 'ttl'
XML_TAG_UPDATED = 'updated'
XML_TAG_URL = 'url'
XML_TAG_WIDTH = 'width'



"""
WA Department of Fire and Emergency Services (DFES) Feed.

Fetches GeoRSS feed from WA Department of Fire and Emergency Services (DFES)
Feed.
"""
import logging

from typing import Optional

from georss_client import GeoRssFeed, FeedEntry
from georss_client.consts import CUSTOM_ATTRIBUTE
from georss_client.exceptions import GeoRssException

_LOGGER = logging.getLogger(__name__)

ADDITIONAL_NAMESPACES = {
    'http://emergency.wa.gov.au/xmlns/dfes': 'dfes'
}

ATTRIBUTION = "Department of Fire and Emergency Services"

REGEXP_ATTR_CATEGORY_WARNINGS = '<b>Category: </b>(?P<{}>[^<]+)</div>'\
    .format(CUSTOM_ATTRIBUTE)
REGEXP_ATTR_CATEGORY_ALL_INCIDENTS = '^(?P<{}>[^<]+) <'\
    .format(CUSTOM_ATTRIBUTE)
REGEXP_ATTR_REGION = '<region>(?P<{}>[^<]+)</region>'\
    .format(CUSTOM_ATTRIBUTE)

URL_PREFIX = 'https://www.emergency.wa.gov.au/data/'
URLS = {
    'warnings': URL_PREFIX + 'message.rss',
    'all_incidents': URL_PREFIX + 'incident_FCAD.rss',
}

XML_TAG_DFES_REGION = 'dfes:region'


class WaDfesFeed(GeoRssFeed):
    """Department of Fire and Emergency Services (DFES) feed."""

    def __init__(self, home_coordinates, feed, filter_radius=None,
                 filter_categories=None):
        """Initialise this service."""
        if feed in URLS:
            super().__init__(home_coordinates, URLS[feed],
                             filter_radius=filter_radius,
                             filter_categories=filter_categories)
            self._feed = feed
        else:
            _LOGGER.error("Unknown feed category %s", feed)
            raise GeoRssException("Feed category must be one of %s".format(
                URLS.keys()))

    def _new_entry(self, home_coordinates, rss_entry, global_data):
        """Generate a new entry."""
        if self._feed == 'warnings':
            return WaDfesWarningsFeedEntry(home_coordinates, rss_entry)
        if self._feed == 'all_incidents':
            return WaDfesAllIncidentsFeedEntry(home_coordinates, rss_entry)

    def _additional_namespaces(self):
        """Provide additional namespaces, relevant for this feed."""
        return ADDITIONAL_NAMESPACES

    def _filter_entries(self, entries):
        """Filter the provided entries."""
        entries = super()._filter_entries(entries)
        if self._filter_categories:
            return list(filter(lambda entry:
                               entry.category in self._filter_categories,
                               entries))
        return entries


class WaDfesFeedEntry(FeedEntry):
    """Department of Fire and Emergency Services (DFES) feed entry."""

    def __init__(self, home_coordinates, rss_entry):
        """Initialise this service."""
        super().__init__(home_coordinates, rss_entry)

    @property
    def attribution(self) -> str:
        """Return the attribution of this entry."""
        return ATTRIBUTION


class WaDfesWarningsFeedEntry(WaDfesFeedEntry):
    """Department of Fire and Emergency Services (DFES) Warnings feed entry."""

    @property
    def category(self) -> str:
        """Return the type of this entry."""
        return self._search_in_description(REGEXP_ATTR_CATEGORY_WARNINGS)

    @property
    def region(self) -> Optional[str]:
        """Return the region of this entry."""
        if self._rss_entry:
            return self._rss_entry.get_additional_attribute(
                XML_TAG_DFES_REGION)
        return None


class WaDfesAllIncidentsFeedEntry(WaDfesFeedEntry):
    """Department of Fire and Emergency Services (DFES) All Incidents feed
    entry."""

    @property
    def category(self) -> str:
        """Return the type of this entry."""
        return self._search_in_description(REGEXP_ATTR_CATEGORY_ALL_INCIDENTS)

    @property
    def region(self) -> str:
        """Return the region of this entry."""
        return self._search_in_description(REGEXP_ATTR_REGION)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

REQUIRES = [
    'haversine>=1.0.1',
    'xmltodict>=0.11.0',
    'requests>=2.20.0',
    'dateparser>=0.7.0',
]

setup(
    name="georss_client",
    version="0.6",
    author="Malte Franken",
    author_email="coding@subspace.de",
    description="A GeoRSS client library.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/exxamalte/python-georss-client",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=REQUIRES
)



"""
GeoRSS feed models.
"""
import datetime
import logging
from typing import Optional

from georss_client.consts import XML_TAG_GEORSS_POLYGON, XML_TAG_GEO_LONG, \
    XML_TAG_GEO_LAT, XML_TAG_GEO_POINT, XML_TAG_GML_POS, XML_TAG_GML_POINT, \
    XML_TAG_GEORSS_WHERE, XML_TAG_GEORSS_POINT, XML_ATTR_TERM, \
    XML_TAG_CATEGORY, XML_TAG_SOURCE, XML_CDATA, XML_TAG_ID, \
    XML_TAG_GUID, XML_TAG_ENTRY, XML_TAG_ITEM, XML_TAG_NAME, XML_TAG_AUTHOR, \
    XML_TAG_LAST_BUILD_DATE, XML_TAG_TTL, XML_TAG_LANGUAGE, \
    XML_TAG_GENERATOR, XML_TAG_COPYRIGHT, XML_TAG_DC_DATE, XML_TAG_PUB_DATE, \
    XML_TAG_PUBLISHED, XML_TAG_UPDATED, XML_TAG_LINK, XML_TAG_CONTENT, \
    XML_TAG_SUMMARY, XML_TAG_DESCRIPTION, XML_TAG_TITLE, XML_TAG_GML_POLYGON, XML_TAG_GML_EXTERIOR, \
    XML_TAG_GML_LINEAR_RING, XML_TAG_GML_POS_LIST, XML_TAG_MANAGING_EDITOR, \
    XML_TAG_CONTRIBUTOR, XML_TAG_RIGHTS, XML_ATTR_HREF, XML_TAG_IMAGE, \
    XML_TAG_URL, XML_TAG_HEIGHT, XML_TAG_WIDTH, XML_TAG_DOCS, XML_TAG_SUBTITLE
from georss_client.xml_parser.geometry import Geometry, Point, Polygon

_LOGGER = logging.getLogger(__name__)


class FeedDictSource:
    """Represents a subset of a feed based on a dict."""

    def __init__(self, source):
        """Initialise feed."""
        self._source = source

    def __repr__(self):
        """Return string representation of this feed item."""
        return '<{}({})>'.format(
            self.__class__.__name__, self.link)

    def _attribute(self, names):
        """Get an attribute from this feed or feed item."""
        if self._source and names:
            # Try each name, and return the first value that is not None.
            for name in names:
                value = self._source.get(name, None)
                if value:
                    return value
        return None

    def _attribute_with_text(self, names):
        """Get an attribute with text from this feed or feed item."""
        value = self._attribute(names)
        if value and isinstance(value, dict) and XML_CDATA in value:
            # <tag attr="/some.uri">Value</tag>
            value = value.get(XML_CDATA)
        return value

    @staticmethod
    def _attribute_in_structure(obj, keys):
        """Return the attribute found under the chain of keys."""
        key = keys.pop(0)
        if key in obj:
            return FeedDictSource._attribute_in_structure(
                obj[key], keys) if keys else obj[key]

    @property
    def title(self) -> Optional[str]:
        """Return the title of this feed or feed item."""
        return self._attribute_with_text([XML_TAG_TITLE])

    @property
    def description(self) -> Optional[str]:
        """Return the description of this feed or feed item."""
        return self._attribute_with_text([XML_TAG_DESCRIPTION,
                                          XML_TAG_SUMMARY,
                                          XML_TAG_CONTENT])

    @property
    def summary(self) -> Optional[str]:
        """Return the summary of this feed or feed item."""
        return self.description

    @property
    def content(self) -> Optional[str]:
        """Return the content of this feed or feed item."""
        return self.description

    @property
    def link(self) -> Optional[str]:
        """Return the link of this feed or feed item."""
        link = self._attribute([XML_TAG_LINK])
        if link and XML_ATTR_HREF in link:
            link = link.get(XML_ATTR_HREF)
        return link

    def get_additional_attribute(self, name):
        """Get an additional attribute not provided as property."""
        return self._attribute([name])


class FeedOrFeedItem(FeedDictSource):
    """Represents the common base of feed and its items."""

    @property
    def category(self) -> Optional[list]:
        """Return the categories of this feed item."""
        category = self._attribute([XML_TAG_CATEGORY])
        if category:
            if isinstance(category, str) or isinstance(category, dict):
                # If it's a string or a dict, wrap in list.
                category = [category]
            result = []
            for item in category:
                if XML_ATTR_TERM in item:
                    # <category term="Category 1"/>
                    item = item.get(XML_ATTR_TERM)
                result.append(item)
            return result
        return None

    @property
    def published_date(self) -> Optional[datetime.datetime]:
        """Return the published date of this feed or feed item."""
        return self._attribute([XML_TAG_PUB_DATE,
                                XML_TAG_PUBLISHED,
                                XML_TAG_DC_DATE])

    @property
    def pub_date(self) -> Optional[datetime.datetime]:
        """Return the published date of this feed or feed item."""
        return self.published_date

    @property
    def updated_date(self) -> Optional[datetime.datetime]:
        """Return the updated date of this feed or feed item."""
        return self._attribute([XML_TAG_LAST_BUILD_DATE, XML_TAG_UPDATED])

    @property
    def last_build_date(self) -> Optional[datetime.datetime]:
        """Return the last build date of this feed."""
        return self.updated_date

    @property
    def author(self) -> Optional[str]:
        """Return the author of this feed."""
        # <managingEditor>jrc-ems@ec.europa.eu</managingEditor>
        managing_editor = self._attribute([XML_TAG_MANAGING_EDITOR])
        if managing_editor:
            return managing_editor
        # <author>
        #   <name>Istituto Nazionale di Geofisica e Vulcanologia</name>
        #   <uri>http://www.ingv.it</uri>
        # </author>
        author = self._attribute([XML_TAG_AUTHOR, XML_TAG_CONTRIBUTOR])
        if author:
            name = author.get(XML_TAG_NAME, None)
            return name
        return None

    @property
    def contributor(self) -> Optional[str]:
        """Return the contributor of this feed."""
        return self.author

    @property
    def managing_editor(self) -> Optional[str]:
        """Return the managing editor of this feed."""
        return self.author


class Feed(FeedOrFeedItem):
    """Represents a feed."""

    @property
    def subtitle(self) -> Optional[str]:
        """Return the subtitle of this feed."""
        return self._attribute_with_text([XML_TAG_SUBTITLE])

    @property
    def copyright(self) -> Optional[str]:
        """Return the copyright of this feed."""
        return self._attribute_with_text([XML_TAG_COPYRIGHT, XML_TAG_RIGHTS])

    @property
    def rights(self) -> Optional[str]:
        """Return the rights of this feed."""
        return self.copyright

    @property
    def generator(self) -> Optional[str]:
        """Return the generator of this feed."""
        return self._attribute_with_text([XML_TAG_GENERATOR])

    @property
    def language(self) -> Optional[str]:
        """Return the language of this feed."""
        return self._attribute([XML_TAG_LANGUAGE])

    @property
    def docs(self) -> Optional[str]:
        """Return the docs URL of this feed."""
        return self._attribute_with_text([XML_TAG_DOCS])

    @property
    def ttl(self) -> Optional[int]:
        """Return the ttl of this feed."""
        return self._attribute([XML_TAG_TTL])

    @property
    def image(self):
        """Return the image of this feed."""
        image = self._attribute([XML_TAG_IMAGE])
        if image:
            return FeedImage(image)
        return None

    @property
    def entries(self):
        """Return the entries of this feed."""
        items = self._attribute([XML_TAG_ITEM, XML_TAG_ENTRY])
        entries = []
        if items and isinstance(items, list):
            for item in items:
                entries.append(FeedItem(item))
        else:
            # A single item in the feed is not represented as an array.
            entries.append(FeedItem(items))
        return entries


class FeedImage(FeedDictSource):
    """Represents a feed image."""

    @property
    def url(self) -> Optional[str]:
        """Return the url of this feed image."""
        return self._attribute([XML_TAG_URL])

    @property
    def height(self) -> Optional[int]:
        """Return the height of this feed image."""
        return self._attribute([XML_TAG_HEIGHT])

    @property
    def width(self) -> Optional[int]:
        """Return the width of this feed image."""
        return self._attribute([XML_TAG_WIDTH])


class FeedItem(FeedOrFeedItem):
    """Represents a feed item."""

    def __repr__(self):
        """Return string representation of this feed item."""
        return '<{}({})>'.format(
            self.__class__.__name__, self.guid)

    @property
    def guid(self) -> Optional[str]:
        """Return the guid of this feed item."""
        return self._attribute_with_text([XML_TAG_GUID, XML_TAG_ID])

    @property
    def id(self) -> Optional[str]:
        """Return the id of this feed item."""
        return self.guid

    @property
    def source(self) -> Optional[str]:
        """Return the source of this feed item."""
        return self._attribute([XML_TAG_SOURCE])

    @property
    def geometry(self) -> Optional[Geometry]:
        """Return the geometry of this feed item."""
        # <georss:point>-0.5 119.8</georss:point>
        point = self._attribute([XML_TAG_GEORSS_POINT])
        if point:
            return Point(point[0], point[1])
        # GML
        where = self._attribute([XML_TAG_GEORSS_WHERE])
        if where:
            # Point:
            # <georss:where>
            #   <gml:Point>
            #     <gml:pos>44.11 -66.23</gml:pos>
            #   </gml:Point>
            # </georss:where>
            pos = self._attribute_in_structure(
                where, [XML_TAG_GML_POINT, XML_TAG_GML_POS])
            if pos:
                return Point(pos[0], pos[1])
            # Polygon:
            # <georss:where>
            #   <gml:Polygon>
            #     <gml:exterior>
            #       <gml:LinearRing>
            #         <gml:posList>
            #           -71.106216 42.366661
            #           -71.105576 42.367104
            #           -71.104378 42.367134
            #           -71.103729 42.366249
            #           -71.098793 42.363331
            #           -71.101028 42.362541
            #           -71.106865 42.366123
            #           -71.106216 42.366661
            #         </gml:posList>
            #       </gml:LinearRing>
            #     </gml:exterior>
            #   </gml:Polygon>
            # </georss:where>
            pos_list = self._attribute_in_structure(
                where, [XML_TAG_GML_POLYGON, XML_TAG_GML_EXTERIOR,
                        XML_TAG_GML_LINEAR_RING, XML_TAG_GML_POS_LIST])
            if pos_list:
                return self._create_polygon(pos_list)
        # <geo:Point xmlns:geo="http://www.w3.org/2003/01/geo/wgs84_pos#">
        #   <geo:lat>38.3728</geo:lat>
        #   <geo:long>15.7213</geo:long>
        # </geo:Point>
        point = self._attribute([XML_TAG_GEO_POINT])
        if point:
            lat = point.get(XML_TAG_GEO_LAT)
            long = point.get(XML_TAG_GEO_LONG)
            if long and lat:
                return Point(lat, long)
        # <geo:long>119.948006</geo:long>
        # <geo:lat>-23.126413</geo:lat>
        lat = self._attribute([XML_TAG_GEO_LAT])
        long = self._attribute([XML_TAG_GEO_LONG])
        if long and lat:
            return Point(lat, long)
        # <georss:polygon>
        #   -34.937663524 148.597260613
        #   -34.9377026399999 148.597169138
        #   -34.9377002169999 148.59708737
        #   -34.9376945989999 148.59705595
        #   -34.9376863529999 148.596955098
        #   -34.937663524 148.597260613
        # </georss:polygon>
        polygon = self._attribute([XML_TAG_GEORSS_POLYGON])
        if polygon:
            # For now, only supporting the first polygon.
            if isinstance(polygon, list) and isinstance(polygon[0], list):
                polygon = polygon[0]
            return self._create_polygon(polygon)
        # None of the above
        return None

    @staticmethod
    def _create_polygon(coordinates):
        """Create a polygon from the provided coordinates."""
        if coordinates:
            if len(coordinates) % 2 != 0:
                # Not even number of coordinates - chop last entry.
                coordinates = coordinates[0:len(coordinates)-1]
            points = []
            for i in range(0, len(coordinates), 2):
                points.append(Point(coordinates[i], coordinates[i + 1]))
            return Polygon(points)
        return None



"""
Natural Resources Canada Earthquakes Feed.

Fetches GeoRSS feed from Natural Resources Canada Earthquakes.
"""
import logging
from datetime import datetime

from typing import Optional

from georss_client import GeoRssFeed, FeedEntry
from georss_client.consts import CUSTOM_ATTRIBUTE
from georss_client.exceptions import GeoRssException

_LOGGER = logging.getLogger(__name__)

ATTRIBUTIONS = {
    'en': "Natural Resources Canada",
    'fr': "Ressources naturelles Canada"
}

REGEXP_ATTR_MAGNITUDE = '<b>magnitude: </b>(?P<{}>[^<]+)<br/>'\
    .format(CUSTOM_ATTRIBUTE)

URL_PATTERN = 'http://www.earthquakescanada.nrcan.gc.ca/index-{}.php?' \
              'tpl_region=canada&tpl_output=rss'
URLS = {
    'en': URL_PATTERN.format('en'),
    'fr': URL_PATTERN.format('fr'),
}


class NaturalResourcesCanadaEarthquakesFeed(GeoRssFeed):
    """Natural Resources Canada Earthquakes feed."""

    def __init__(self, home_coordinates, language, filter_radius=None,
                 filter_minimum_magnitude=None):
        """Initialise this service."""
        if language in URLS:
            super().__init__(home_coordinates, URLS[language],
                             filter_radius=filter_radius)
            self._language = language
            self._filter_minimum_magnitude = filter_minimum_magnitude
        else:
            _LOGGER.error("Unknown feed language %s", language)
            raise GeoRssException("Feed language must be one of %s".format(
                URLS.keys()))

    def __repr__(self):
        """Return string representation of this feed."""
        return '<{}(home={}, url={}, radius={}, magnitude={})>'.format(
            self.__class__.__name__, self._home_coordinates, self._url,
            self._filter_radius, self._filter_minimum_magnitude)

    def _new_entry(self, home_coordinates, rss_entry, global_data):
        """Generate a new entry."""
        return NaturalResourcesCanadaEarthquakesFeedEntry(
            home_coordinates, ATTRIBUTIONS[self._language], rss_entry)

    def _filter_entries(self, entries):
        """Filter the provided entries."""
        entries = super()._filter_entries(entries)
        if self._filter_minimum_magnitude:
            # Return only entries that have an actual magnitude value, and
            # the value is equal or above the defined threshold.
            return list(filter(lambda entry:
                               entry.magnitude and entry.magnitude >= self.
                               _filter_minimum_magnitude, entries))
        return entries


class NaturalResourcesCanadaEarthquakesFeedEntry(FeedEntry):
    """Natural Resources Canada Earthquakes feed entry."""

    def __init__(self, home_coordinates, attribution, rss_entry):
        """Initialise this service."""
        super().__init__(home_coordinates, rss_entry)
        self._attribution = attribution

    @property
    def attribution(self) -> str:
        """Return the attribution of this entry."""
        return self._attribution

    @property
    def magnitude(self) -> Optional[float]:
        """Return the type of this entry."""
        magnitude = self._search_in_description(REGEXP_ATTR_MAGNITUDE)
        if magnitude:
            # Convert to float. In the 'fr' version of the feed, the
            # magnitude value uses comma as decimal separator, hence replacing
            # comma with dot here. Because magnitude is never above 1000 the
            # value in the feed will never have a comma as thousands-separator
            # in the 'en' version of the feed.
            magnitude = float(magnitude.replace(',', '.'))
        return magnitude
